#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""DICOM schema for Image model."""

from dataclasses import dataclass
from typing import Any, Dict, Optional, Sequence, Type

from marshmallow import fields, post_load, pre_dump
from pydicom.dataset import Dataset

from wsidicom.codec import LossyCompressionIsoStandard
from wsidicom.geometry import SizeMm
from wsidicom.metadata.image import (
    ExtendedDepthOfField,
    FocusMethod,
    Image,
    ImageCoordinateSystem,
    LossyCompression,
)
from wsidicom.metadata.schema.dicom.defaults import Defaults
from wsidicom.metadata.schema.dicom.fields import (
    BooleanDicomField,
    DateTimeDicomField,
    DefaultingDicomField,
    EnumDicomField,
    FlattenOnDumpNestedDicomField,
    FloatDicomField,
    ImageOrientationSlideField,
    ListDicomField,
    NestedDatasetDicomField,
    OffsetInSlideCoordinateSystemField,
    PixelSpacingDicomField,
    StringDicomField,
)
from wsidicom.metadata.schema.dicom.schema import (
    DicomSchema,
    ModuleDicomSchema,
)


class ExtendedDepthOfFieldDicomSchema(DicomSchema[ExtendedDepthOfField]):
    number_of_focal_planes = fields.Integer(
        data_key="NumberOfFocalPlanes", allow_none=False
    )
    distance_between_focal_planes = fields.Float(
        data_key="DistanceBetweenFocalPlanes", allow_none=False
    )

    @property
    def load_type(self) -> Type[ExtendedDepthOfField]:
        return ExtendedDepthOfField


class ImageCoordinateSystemDicomSchema(DicomSchema[ImageCoordinateSystem]):
    origin = OffsetInSlideCoordinateSystemField(
        data_key="TotalPixelMatrixOriginSequence",
        allow_none=False,
        dump_default=Defaults.image_coordinate_system_origin,
    )
    rotation = ImageOrientationSlideField(
        data_key="ImageOrientationSlide",
        allow_none=False,
        dump_default=Defaults.image_coordinate_system_rotation,
    )

    @property
    def load_type(self) -> Type[ImageCoordinateSystem]:
        return ImageCoordinateSystem

    def load(self, dataset: Dataset, **kwargs) -> Optional[ImageCoordinateSystem]:
        try:
            return super().load(dataset, **kwargs)
        except TypeError:
            return None


@dataclass(frozen=True)
class PixelMeasureDicomModel:
    pixel_spacing: Optional[SizeMm] = None
    focal_plane_spacing: Optional[float] = None
    depth_of_field: Optional[float] = None


class PixelMeasureDicomSchema(DicomSchema[PixelMeasureDicomModel]):
    pixel_spacing = PixelSpacingDicomField(data_key="PixelSpacing", allow_none=True)
    focal_plane_spacing = FloatDicomField(
        data_key="SpacingBetweenSlices", allow_none=True
    )
    depth_of_field = FloatDicomField(data_key="SliceThickness", allow_none=True)

    @property
    def load_type(self) -> Type[PixelMeasureDicomModel]:
        return PixelMeasureDicomModel


class LossyCompressionDicomSchema:
    method = StringDicomField()
    ratio = FloatDicomField()


class LossyCompressionsDicomSchema(DicomSchema[Sequence[LossyCompression]]):
    methods = ListDicomField(
        EnumDicomField(LossyCompressionIsoStandard, by_value=True),
        data_key="LossyImageCompressionMethod",
        dump_none_if_empty=True,
    )
    ratios = ListDicomField(
        FloatDicomField(),
        data_key="LossyImageCompressionRatio",
        dump_none_if_empty=True,
    )
    lossy_compressed = BooleanDicomField(
        data_key="LossyImageCompression", dump_only=True, truthy="01", falsy="00"
    )

    @property
    def load_type(self) -> Type[Sequence[LossyCompression]]:
        return list

    @pre_dump
    def pre_dump(
        self, lossy_compressions: Sequence[LossyCompression], **kwargs
    ) -> Dict[str, Any]:
        return {
            "methods": [compression.method for compression in lossy_compressions],
            "ratios": [compression.ratio for compression in lossy_compressions],
            "lossy_compressed": len(lossy_compressions) > 0,
        }

    @post_load
    def post_load(self, data: Dict[str, Any], **kwargs) -> Sequence[LossyCompression]:
        methods = data.pop("methods", [])
        ratios = data.pop("ratios", [])
        if len(methods) != len(ratios):
            raise ValueError(
                (
                    f"Number of lossy compression methods {len(methods)} did not match "
                    f"number of ratios {len(ratios)}."
                )
            )
        return [
            LossyCompression(method, ratio) for method, ratio in zip(methods, ratios)
        ]


class ImageDicomSchema(ModuleDicomSchema[Image]):
    acquisition_datetime = DefaultingDicomField(
        DateTimeDicomField(),
        data_key="AcquisitionDateTime",
        dump_default=Defaults.date_time,
        load_default=None,
    )
    focus_method = DefaultingDicomField(
        fields.Enum(FocusMethod),
        data_key="FocusMethod",
        dump_default=Defaults.focus_method,
        load_default=None,
    )
    extended_depth_of_field_bool = BooleanDicomField(
        data_key="ExtendedDepthOfField", load_default=False
    )
    extended_depth_of_field = FlattenOnDumpNestedDicomField(
        ExtendedDepthOfFieldDicomSchema(),
        allow_none=True,
        load_default=None,
    )
    image_coordinate_system = FlattenOnDumpNestedDicomField(
        ImageCoordinateSystemDicomSchema(),
        allow_none=True,
        load_default=None,
    )
    pixel_measure = NestedDatasetDicomField(
        PixelMeasureDicomSchema(),
        data_key="SharedFunctionalGroupsSequence",
        nested_data_key="PixelMeasuresSequence",
    )
    lossy_compressions = FlattenOnDumpNestedDicomField(
        LossyCompressionsDicomSchema(),
        allow_none=True,
        load_default=None,
    )

    @property
    def load_type(self) -> Type[Image]:
        return Image

    @pre_dump
    def pre_dump(self, image: Image, **kwargs):
        return {
            "acquisition_datetime": image.acquisition_datetime,
            "focus_method": image.focus_method,
            "extended_depth_of_field_bool": image.extended_depth_of_field is not None,
            "extended_depth_of_field": image.extended_depth_of_field,
            "image_coordinate_system": image.image_coordinate_system,
            "pixel_measure": PixelMeasureDicomModel(
                pixel_spacing=image.pixel_spacing,
                focal_plane_spacing=image.focal_plane_spacing,
                depth_of_field=image.depth_of_field,
            ),
            "lossy_compressions": image.lossy_compressions
            if image.lossy_compressions
            else [],
        }

    @post_load
    def post_load(self, data: Dict[str, Any], **kwargs):
        extended_depth_of_field_bool = data.pop("extended_depth_of_field_bool")
        extended_depth_of_field = data.get("extended_depth_of_field", None)
        if (extended_depth_of_field_bool) != (extended_depth_of_field is not None):
            raise ValueError(
                (
                    f"Extended depth of field bool {extended_depth_of_field_bool} did ",
                    f"not match depth of field data {extended_depth_of_field}.",
                )
            )
        pixel_measure: Optional[PixelMeasureDicomModel] = data.pop(
            "pixel_measure", None
        )
        if pixel_measure is not None:
            data["pixel_spacing"] = pixel_measure.pixel_spacing
            data["focal_plane_spacing"] = pixel_measure.focal_plane_spacing
            data["depth_of_field"] = pixel_measure.depth_of_field

        return super().post_load(data, **kwargs)

    @property
    def module_name(self) -> str:
        return "image"



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module with base IO class for handling DICOM WSI files."""

import struct
from datetime import datetime
from pathlib import Path
from struct import pack
from typing import Any, BinaryIO, Callable, Dict, Literal, Optional, Union

from pydicom.dataelem import DataElement_from_raw, RawDataElement
from pydicom.dataset import Dataset, FileMetaDataset, validate_file_meta
from pydicom.filebase import DicomIO
from pydicom.filereader import (
    _read_file_meta_info,
    data_element_generator,
    read_partial,
    read_preamble,
)
from pydicom.filewriter import write_dataset, write_file_meta_info, writers
from pydicom.tag import BaseTag, SequenceDelimiterTag, Tag
from pydicom.uid import UID
from pydicom.valuerep import VR

from wsidicom.errors import WsiDicomFileError
from wsidicom.tags import ExtendedOffsetTableTag


class WsiDicomIO(DicomIO):
    """Class for reading or writing DICOM WSI to stream."""

    def __init__(
        self,
        stream: BinaryIO,
        little_endian: bool = True,
        implicit_vr: bool = False,
        filepath: Optional[Path] = None,
        owned: bool = False,
    ):
        """
        Create a WsiDicomIO.

        Parameters
        ----------
        stream: BinaryIO
            Stream to use.
        little_endian: bool = True
            If to set the stream to little endian.
        implicit_vr: bool = False
            If to set the stream to implicit VR.
        filepath: Optional[Path] = None
            Optional filepath of stream.
        owned: bool = False
            If the stream should be closed by this instance.
        """
        stream.seek(0)
        self._stream = stream
        self._filepath = filepath
        self._owned = owned
        super().__init__(stream)
        self.is_little_endian = little_endian
        self.is_implicit_VR = implicit_vr
        self.__enter__()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @classmethod
    def open(
        cls,
        filepath: Path,
        mode: Union[Literal["rb"], Literal["r+b"], Literal["w+b"]],
        little_endian: bool = True,
        implicit_vr: bool = False,
    ) -> "WsiDicomIO":
        """Open a file and return a WsiDicomIO instance.

        Parameters
        ----------
        filepath: Path
            Path to file.
        mode: Union[Literal["rb"], Literal["w+b"]]
            Mode to open file in.
        little_endian: bool = True
            If to set the stream to little endian.
        implicit_vr: bool = False
            If to set the stream to implicit VR.

        Returns
        -------
        WsiDicomIO
            Instance of WsiDicomIO.
        """
        stream = open(filepath, mode)
        return cls(stream, little_endian, implicit_vr, filepath=filepath, owned=True)

    @property
    def owned(self) -> bool:
        """Return True if the stream is owned by this instance."""
        return self._owned

    @property
    def closed(self) -> bool:
        """Return True if the stream is closed."""
        return self._stream.closed

    @property
    def filepath(self) -> Optional[Path]:
        """Return filepath, if opened from file."""
        return self._filepath

    @property
    def write(self) -> Callable[[bytes], int]:
        return self._stream.write

    @property
    def seek(self):
        return self._stream.seek

    @property
    def tell(self):
        return self._stream.tell

    @property
    def parent_read(self):
        return self._stream.read

    def read_media_storage_sop_class_uid(self) -> UID:
        """Read Media Storage SOP Class UID from file meta info."""
        metadata = self.read_file_meta_info()
        self.seek(0)
        return metadata.MediaStorageSOPClassUID

    def read_file_meta_info(self) -> FileMetaDataset:
        """Read file meta info from stream."""
        self.seek(0)
        read_preamble(self._stream, False)
        return _read_file_meta_info(self._stream)

    def read_dataset(self, force: bool = False) -> Dataset:
        """Read dataset, excluding EOT and PixelData from stream."""
        extended_offset_table_tag = ExtendedOffsetTableTag

        def _stop_at(tag: BaseTag, vr: Optional[str], length: int) -> bool:
            return tag >= extended_offset_table_tag

        self.seek(0)
        return read_partial(
            self._stream,
            _stop_at,
            defer_size=None,
            force=force,
            specific_tags=None,
        )

    def read_tag_length(self, long: bool) -> int:
        """Read tag length."""
        if not long and not self.is_implicit_VR:
            return self.read_US()
        return self.read_UL()

    def read_tag_vr(self) -> Optional[bytes]:
        """Read tag VR if implicit VR."""
        if not self.is_implicit_VR:
            vr = self.read(4, need_exact_length=True)
            return vr[0:2]

    def check_tag_and_length(
        self, tag: BaseTag, length: int, with_vr: bool, long: bool
    ) -> None:
        """Check if tag at position is expected tag with expected length.

        Parameters
        ----------
        tag: BaseTag
            Expected tag.
        length: int
            Expected length.
        with_vr: bool
            If tag is expected to have VR.
        long: bool
            If length is expected to be long.

        """
        try:
            read_tag = self.read_tag()
            if tag != read_tag:
                raise WsiDicomFileError(
                    str(self), f"Found tag {read_tag} expected {tag}."
                )
            if with_vr:
                if self.is_implicit_VR:
                    raise WsiDicomFileError(str(self), "Expected VR, but implicit VR.")
                self.read_tag_vr()
            read_length = self.read_tag_length(long)
            if length != read_length:
                raise WsiDicomFileError(
                    str(self), f"Found length {read_length} expected {length}."
                )
        except struct.error:
            raise WsiDicomFileError(str(self), "Failed to unpack data.")

    def read_sequence_delimiter(self):
        """Check if last read tag was a sequence delimiter.
        Raises WsiDicomFileError otherwise.
        """
        TAG_BYTES = 4
        self.seek(-TAG_BYTES, 1)
        if self.read_le_tag() != SequenceDelimiterTag:
            raise WsiDicomFileError(str(self), "No sequence delimiter tag")

    def write_unsigned_long_long(self, value: int):
        """Write unsigned long long integer (64 bits).

        Parameters
        ----------
        value: int
            Value to write.

        """
        if self.is_little_endian:
            format = "<Q"
        else:
            format = ">Q"
        self.write(pack(format, value))

    def write_tag_of_vr_and_length(
        self, tag: BaseTag, value_representation: str, length: Optional[int] = None
    ):
        """Write tag, tag VR and length.

        Parameters
        ----------
        tag: str
            Name of tag to write.
        value_representation: str.
            Value representation (VR) of tag to write.
        length: Optional[int] = None
            Length of data after tag. 'Unspecified' (0xFFFFFFFF) if None.

        """
        if self.is_little_endian:
            write_ul = self.write_leUL
            write_us = self.write_leUS
        else:
            write_ul = self.write_beUL
            write_us = self.write_beUS
        self.write_tag(Tag(tag))
        if not self.is_implicit_VR:
            self.write(bytes(value_representation, "iso8859"))
            write_us(0)
        if length is not None:
            write_ul(length)
        else:
            write_ul(0xFFFFFFFF)

    def write_preamble(self):
        """Write DICOM preamble."""
        self.seek(0)
        preamble = b"\x00" * 128
        self.write(preamble)
        self.write(b"DICM")

    def write_file_meta_info(
        self, instance_uid: UID, sop_class_uid: UID, transfer_syntax: UID
    ):
        """Write file meta info.

        Parameters
        ----------
        instance_uid: UID
            SOP Instance UID.
        sop_class_uid: UID
            SOP Class UID.

        """
        meta = FileMetaDataset()
        meta.TransferSyntaxUID = transfer_syntax
        meta.MediaStorageSOPInstanceUID = instance_uid
        meta.MediaStorageSOPClassUID = sop_class_uid
        validate_file_meta(meta)
        write_file_meta_info(self, meta)

    def write_dataset(self, dataset: Dataset, content_datetime: datetime):
        """Write dataset to stream.

        Parameters
        ----------
        dataset: Dataset
            Dataset to write.

        """
        dataset.ContentDate = datetime.date(content_datetime).strftime("%Y%m%d")
        dataset.ContentTime = datetime.time(content_datetime).strftime("%H%M%S.%f")
        write_dataset(self, dataset)

    def close(self, force: Optional[bool] = False) -> None:
        """Close stream if owned by instance or forced."""
        if self._owned or force:
            self._stream.close()

    def update_dataset(self, dataset_start: int, update: Dict[BaseTag, Any]):
        """Update dataset in place.

        The element value representation must allow padding,
        and values to replace should be padded to the same length or longer than the
        replacement value.


        Parameters
        ----------
        dataset_start: int
            Position of dataset start.
        update: Dict[BaseTag, Any]
            Dictionary with tags to update and their new values.
        """

        rewind = self.tell()
        self.seek(dataset_start)
        for element in data_element_generator(
            self._stream,
            self.is_implicit_VR,
            self.is_little_endian,
            specific_tags=list(update.keys()),
        ):
            if element.tag not in update.keys():
                # generator can include `Specific Character Set` element
                continue
            if not isinstance(element, RawDataElement):
                raise ValueError("Can only update raw data elements.")
            element_value_position = element.value_tell
            length = element.length
            if isinstance(element, RawDataElement):
                element = DataElement_from_raw(element)
            element.value = update[BaseTag(element.tag)]
            self.seek(element_value_position)
            writer, param = writers[VR(element.VR)]
            if param is None:
                param = []
            writer(self, element, *param)
            if self.tell() > element_value_position + length:
                raise ValueError("Updated element is longer than original.")
        self.seek(rewind)



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import Iterable, List, Optional, OrderedDict

from wsidicom.errors import (
    WsiDicomNotFoundError,
    WsiDicomOutOfBoundsError,
)
from wsidicom.geometry import Size, SizeMm
from wsidicom.group import Level
from wsidicom.group.level import BaseLevel
from wsidicom.instance import ImageType, WsiInstance
from wsidicom.metadata import ImageCoordinateSystem
from wsidicom.series.series import Series
from wsidicom.stringprinting import list_pretty_str


class Pyramid(Series[Level]):
    """Represents a list of Levels of the volume (e.g. pyramidal level) wsi flavor
    forming a WSI pyramid. All levels in the pyramid must have the same image origin
    and extended depth of field."""

    def __init__(self, levels: Iterable[Level]):
        """Holds a stack of levels.

        Parameters
        ----------
        levels: Iterable[Level]
            List of levels to include in the pyramid.
        """
        self._levels = OrderedDict(
            (level.level, level)
            for level in sorted(levels, key=lambda level: level.level)
        )
        self._groups = list(self._levels.values())
        if len(self._levels) != 0 and self._levels[0].uids is not None:
            self._uids = self._validate_series(list(self._levels.values()))
        else:
            self._uids = None
        mm_size = next(
            level.default_instance.mm_size
            for level in self._levels.values()
            if level.default_instance.mm_size is not None
        )
        if mm_size is None:
            raise ValueError(
                "ImagedVolumeWidth and ImagedVolumeHeight must be set for "
                '"Volume" type'
            )
        self._mm_size = mm_size

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._levels})"

    def __str__(self) -> str:
        """Return string representation of pyramid."""
        return self.pretty_str()

    def __getitem__(self, index: int) -> Level:
        return list(self._levels.values())[index]

    def __len__(self) -> int:
        return len(self._levels)

    def get(self, index: int, pyramid_index: bool = True) -> Level:
        """Get level by index.

        Parameters
        ----------
        index: int
            Index in pyramid to get
        pyramid_index: bool = True
            If to get level by pyramid index (True) or list index (False).

        Returns
        -------
        Level
            The level at index in the pyramid
        """
        try:
            if pyramid_index:
                return self._levels[index]
            return self[index]
        except (KeyError, IndexError):
            raise WsiDicomNotFoundError(f"Level index {index}", "pyramid")

    @property
    def image_type(self) -> ImageType:
        return ImageType.VOLUME

    @property
    def size(self) -> Size:
        return self.base_level.size

    @property
    def pixel_spacing(self) -> SizeMm:
        return self.base_level.pixel_spacing

    @property
    def mpp(self) -> SizeMm:
        return self.base_level.mpp

    @property
    def tile_size(self) -> Size:
        return self.base_level.tile_size

    @classmethod
    def open(cls, instances: Iterable[WsiInstance]) -> "Pyramid":
        """Return pyramid created from wsi instances.

        Parameters
        ----------
        instances: Iterable[WsiInstance]
            Instances to create pyramid from.

        Returns
        -------
        Pyramid
            Created pyramid.
        """
        instances_grouped_by_size = cls._group_instances_by_size(instances)
        base_level = BaseLevel(next(instances_grouped_by_size))
        levels: List[Level] = [base_level]
        levels.extend(
            [
                Level(instances, base_level.pixel_spacing)
                for instances in instances_grouped_by_size
            ]
        )
        return cls(levels)

    def pretty_str(self, indent: int = 0, depth: Optional[int] = None) -> str:
        """Return string representation of pyramid."""
        string = self.__class__.__name__
        if depth is not None:
            depth -= 1
            if depth < 0:
                return string
        return (
            string
            + " of levels:\n"
            + list_pretty_str(self._levels.values(), indent + 2, depth)
        )

    @property
    def levels(self) -> List[Level]:
        """Return contained levels."""
        return list(self._levels.values())

    @property
    def pyramid_indices(self) -> List[int]:
        """Return contained levels."""
        return list(self._levels.keys())

    @property
    def highest_level(self) -> int:
        """Return highest valid pyramid level (which results in a 1x1 image)."""
        return self.base_level.get_highest_level()

    @property
    def lowest_single_tile_level(self) -> int:
        """Return lowest pyramid level that has consists of a single tile."""
        return self.base_level.get_lowest_single_tile_level()

    @property
    def base_level(self) -> Level:
        """Return the base level of the pyramid."""
        return self._levels[0]

    @property
    def mm_size(self) -> SizeMm:
        return self._mm_size

    @property
    def image_coordinate_system(self) -> Optional[ImageCoordinateSystem]:
        return self.base_level.image_coordinate_system

    def valid_level(self, level: int) -> bool:
        """Check that given level is less or equal to the highest level
        (1x1 pixel level).

        Parameters
        ----------
        level: int
            The level to check

        Returns
        -------
        bool
            True if level is valid
        """
        return level <= self.highest_level

    def get_closest_by_level(self, level: int) -> Level:
        """Search for level that is closest to and smaller than the given
        level.

        Parameters
        ----------
        level: int
            The level to search for

        Returns
        -------
        Level
            The level closest to searched level
        """
        if not self.valid_level(level):
            raise WsiDicomOutOfBoundsError(
                f"Level {level}", f"maximum level {self.highest_level}"
            )
        closest_level = 0
        closest = None
        for wsi_level in self._levels.values():
            if (level >= wsi_level.level) and (closest_level <= wsi_level.level):
                closest_level = wsi_level.level
                closest = wsi_level
        if closest is None:
            raise WsiDicomNotFoundError(f"Level for {level}", "level series")
        return closest

    def get_closest_by_size(self, size: Size) -> Level:
        """Search for level that by size is closest to and larger than the
        given size.

        Parameters
        ----------
        size: Size
            The size to search for

        Returns
        -------
        Level
            The level with size closest to searched size
        """
        closest_size = self._levels[0].size
        closest = None
        for wsi_level in self._levels.values():
            if (size.width <= wsi_level.size.width) and (
                wsi_level.size.width <= closest_size.width
            ):
                closest_size = wsi_level.size
                closest = wsi_level
        if closest is None:
            raise WsiDicomNotFoundError(f"Level for size {size}", "level series")
        return closest

    def get_closest_by_pixel_spacing(self, pixel_spacing: SizeMm) -> Level:
        """Search for level that by pixel spacing is closest to and smaller
        than the given pixel spacing. Only the spacing in x-axis is used.

        Parameters
        ----------
        pixel_spacing: SizeMm
            Pixel spacing to search for

        Returns
        -------
        Level
            The level with pixel spacing closest to searched spacing
        """
        closest_pixel_spacing: float = 0
        closest = None
        for wsi_level in self._levels.values():
            if (pixel_spacing.width >= wsi_level.pixel_spacing.width) and (
                closest_pixel_spacing <= wsi_level.pixel_spacing.width
            ):
                closest_pixel_spacing = wsi_level.pixel_spacing.width
                closest = wsi_level
        if closest is None:
            raise WsiDicomNotFoundError(
                f"Level for pixel spacing {pixel_spacing}", "level series"
            )
        return closest



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module for reading DICOM WSI files."""

import threading
from functools import cached_property
from pathlib import Path
from typing import List, Optional, Tuple

from pydicom.errors import InvalidDicomError
from pydicom.tag import Tag
from pydicom.uid import UID

from wsidicom.codec import Codec
from wsidicom.errors import WsiDicomFileError, WsiDicomNotSupportedError
from wsidicom.file.io.frame_index import (
    Bot,
    EmptyBot,
    EmptyBotException,
    Eot,
    FrameIndex,
    NativePixelData,
    OffsetTableType,
)
from wsidicom.file.io.wsidicom_io import WsiDicomIO
from wsidicom.instance import ImageType, WsiDataset
from wsidicom.tags import ExtendedOffsetTableTag
from wsidicom.uid import FileUids


class WsiDicomReader:
    """Reader for DICOM WSI data in stream"""

    def __init__(self, stream: WsiDicomIO):
        """
        Parse DICOM stream. If valid WSI type read required parameters.

        Parameters
        ----------
        stream: WsiDicomIO
            File to open.
        """
        self._lock = threading.Lock()
        self._stream = stream
        try:
            file_meta = self._stream.read_file_meta_info()
        except InvalidDicomError:
            raise WsiDicomFileError(str(stream), "is not a DICOM file or stream.")
        self._transfer_syntax_uid = UID(file_meta.TransferSyntaxUID)
        self._stream.is_little_endian = self._transfer_syntax_uid.is_little_endian
        self._stream.is_implicit_VR = self._transfer_syntax_uid.is_implicit_VR
        dataset = self._stream.read_dataset()
        self._pixel_data_position = self._stream.tell()

        self._image_type = WsiDataset.is_supported_wsi_dicom(dataset)
        if self._image_type is not None:
            self._dataset = WsiDataset(dataset)
        else:
            raise WsiDicomNotSupportedError(
                f"Non-supported file or stream {self._stream}."
            )
        syntax_supported = Codec.is_supported(
            self.transfer_syntax,
            self._dataset.samples_per_pixel,
            self._dataset.bits,
            self._dataset.photometric_interpretation,
        )
        if not syntax_supported:
            raise WsiDicomNotSupportedError(
                f"Non-supported transfer syntax {self.transfer_syntax}"
            )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def offset_table_type(self) -> OffsetTableType:
        """Return type of the offset table, or None if not present."""
        return self.frame_index.offset_table_type

    @cached_property
    def frame_index(self) -> FrameIndex:
        return self._get_frame_index()

    @property
    def dataset(self) -> WsiDataset:
        """Return pydicom dataset of stream."""
        return self._dataset

    @property
    def image_type(self) -> Optional[ImageType]:
        return self._image_type

    @property
    def uids(self) -> FileUids:
        """Return uids."""
        return self.dataset.uids

    @property
    def transfer_syntax(self) -> UID:
        """Return transfer syntax uid."""
        return self._transfer_syntax_uid

    @property
    def frame_offset(self) -> int:
        """Return frame offset (for concatenated stream, 0 otherwise)."""
        return self.dataset.frame_offset

    @property
    def frame_positions(self) -> List[Tuple[int, int]]:
        """Return frame positions and lengths."""
        return self.frame_index.index

    @property
    def frame_count(self) -> int:
        """Return number of frames."""
        return self.dataset.frame_count

    @property
    def filepath(self) -> Optional[Path]:
        """Return filename if stream is file."""
        return self._stream.filepath

    def read_frame(self, frame_index: int) -> bytes:
        """Return frame data from pixel data by frame index.

        Parameters
        ----------
        frame_index: int
            Frame, including concatenation offset, to get.

        Returns
        -------
        bytes
            The frame as bytes
        """
        frame_index -= self.frame_offset
        frame_position, frame_length = self.frame_positions[frame_index]
        with self._lock:
            self._stream.seek(frame_position, 0)
            return self._stream.read(frame_length)

    @classmethod
    def open(cls, file: Path) -> "WsiDicomReader":
        """Open file in path as WsiDicomFileReader.

        Parameters
        ----------
        file: Path
            Path to file.

        Returns
        -------
        WsiDicomFileReader
            WsiDicomFileReader for file.
        """
        stream = WsiDicomIO.open(file, "rb")
        return cls(stream)

    def _get_frame_index(self) -> FrameIndex:
        """Create frame index for stream."""
        self._stream.seek(self._pixel_data_position)
        if not self.transfer_syntax.is_encapsulated:
            return NativePixelData(
                self._stream,
                self._pixel_data_position,
                self._dataset.frame_count,
                self._dataset.tile_size,
                self._dataset.samples_per_pixel,
                self._dataset.bits,
            )
        pixel_data_or_eot_tag = Tag(self._stream.read_tag())
        if pixel_data_or_eot_tag == ExtendedOffsetTableTag:
            return Eot(self._stream, self._pixel_data_position, self.frame_count)
        try:
            return Bot(self._stream, self._pixel_data_position, self.frame_count)
        except EmptyBotException:
            self._stream.seek(self._pixel_data_position)
            return EmptyBot(self._stream, self._pixel_data_position, self.frame_count)

    def close(self, force: Optional[bool] = False) -> None:
        """Close stream."""
        self._stream.close(force)



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module for handling DICOM RLE data with imagecodecs PackBits codec."""

import io
from struct import pack
from typing import List, Sequence

import numpy as np
from wsidicom.codec.optionals import (
    packbits_encode,
    IMAGE_CODECS_AVAILABLE,
    dicomrle_decode,
)


class RleCodec:
    """Codec for encoding and decoding DICOM RLE data with imagecodecs PackBits codec."""

    _bits_to_compnents_per_channel = {
        8: 1,
        16: 2,
    }
    _data_type_to_components_per_channel = {
        np.dtype(np.uint8): 1,
        np.dtype(np.uint16): 2,
    }

    @classmethod
    def is_available(cls) -> bool:
        """Return true if codec is available.

        Returns
        -------
        bool
            True if codec is available.
        """
        return IMAGE_CODECS_AVAILABLE

    @classmethod
    def encode(cls, data: np.ndarray) -> bytes:
        """Encode data to DICOM RLE.

        Parameters
        ----------
        data : np.ndarray
            Data to encode. Must be 2D or 3D array with uint8 or uint16 dtype.

        Returns
        -------
        bytes
            Encoded data. First 64 bytes is the RLE header, followed by RLE segments."""
        if not cls.is_available():
            raise RuntimeError("Image codecs not available.")
        components_per_channel = cls._data_type_to_components_per_channel[data.dtype]
        rows = data.shape[0]
        cols = data.shape[1]
        channel_shape = (rows, cols, components_per_channel)
        # Reshape so that there is channels to iterate over
        if data.ndim == 2:
            data = data.reshape((data.shape[0], data.shape[1], 1))
        channel_count = data.shape[2]
        segment_positions: List[int] = []
        with io.BytesIO() as buffer:
            # For each channel
            for channel_index in range(channel_count):
                channel = data[:, :, channel_index]
                if channel.dtype != np.uint8:
                    # Split into high and low byte components
                    channel = channel.copy().view(np.uint8)
                # Reshape so that there is byte component to iterate over
                channel = channel.reshape(channel_shape)
                # Encode each byte component in reverse (high first)
                for component_index in reversed(range(components_per_channel)):
                    segment_positions.append(64 + buffer.tell())
                    component = channel[:, :, component_index]
                    encoded_component = packbits_encode(component)
                    buffer.write(encoded_component)
                    if len(encoded_component) % 2 == 1:
                        # Segment must be even length
                        buffer.write(b"\x00")

            encoded = buffer.getvalue()
        image_codec_header = cls._create_header(segment_positions)
        return image_codec_header + encoded

    @classmethod
    def decode(cls, data: bytes, rows: int, cols: int, bits: int) -> np.ndarray:
        """Decode DICOM RLE data.

        Parameters
        ----------
        data : bytes
            RLE data to decode. Should have a RLE header followed by RLE segments.
        rows : int
            Number of rows in decoded data.
        cols : int
            Number of columns in decoded data.
        bits : int
            Number of bits per pixel. Must be 8 or 16.
        """
        if not cls.is_available():
            raise RuntimeError("Image codecs not available.")
        if bits == 16:
            dtype = np.uint16
        elif bits == 8:
            dtype = np.uint8
        else:
            raise ValueError("Invalid bits.")
        decoded_bytes = dicomrle_decode(data, dtype)
        # Read data is uint8
        decoded = np.frombuffer(decoded_bytes, dtype=dtype)
        # Reshape to channels, rows, cols
        decoded = decoded.reshape(len(decoded) // (rows * cols), rows, cols)
        # Transpose to rows, cols, channels, and squeeze channels if only 1
        decoded = decoded.transpose((1, 2, 0)).squeeze()
        return decoded

    @staticmethod
    def _create_header(segment_positions: Sequence[int]):
        """Create RLE header from segment positions."""
        if len(segment_positions) < 1:
            raise ValueError("Too few segments.")
        if len(segment_positions) > 15:
            raise ValueError("Too many segments.")
        if any(segment % 2 == 1 for segment in segment_positions):
            raise ValueError("Segment positions must be even.")
        segment_positions = list(segment_positions)
        segment_count = len(segment_positions)
        empty_segment_count = 15 - segment_count
        segment_positions.extend([0 for _ in range(empty_segment_count)])
        return pack(
            "<16L",
            segment_count,
            *segment_positions,
        )



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Type,
    Union,
)

from marshmallow import Schema, fields, pre_dump

from wsidicom.conceptcode import (
    AnatomicPathologySpecimenTypesCode,
    ContainerTypeCode,
    SpecimenCollectionProcedureCode,
    SpecimenEmbeddingMediaCode,
    SpecimenFixativesCode,
    SpecimenPreparationStepsCode,
    SpecimenSamplingProcedureCode,
    SpecimenStainsCode,
)
from wsidicom.metadata.sample import (
    BaseSpecimen,
    Collection,
    Embedding,
    Fixation,
    PreparationStep,
    Processing,
    Receiving,
    Sample,
    SampledSpecimen,
    SampleLocalization,
    Sampling,
    SamplingLocation,
    SlideSample,
    Specimen,
    SpecimenIdentifier,
    Staining,
    Storage,
    UnknownSampling,
)
from wsidicom.metadata.schema.common import DataclassLoadingSchema, LoadingSchema
from wsidicom.metadata.schema.json.fields import (
    CodeJsonField,
    JsonFieldFactory,
    MeasurementJsonField,
    SpecimenIdentifierJsonField,
    UidJsonField,
)
from wsidicom.metadata.schema.json.sample.model import (
    BaseSpecimenJsonModel,
    PreparationAction,
    SampleJsonModel,
    SamplingConstraintJsonModel,
    SamplingJsonModel,
    SlideSampleJsonModel,
    SpecimenJsonModel,
)
from wsidicom.metadata.schema.json.sample.parser import SpecimenJsonParser


class SamplingLocationJsonSchema(LoadingSchema[SamplingLocation]):
    reference = fields.String(allow_none=True)
    description = fields.String(allow_none=True)
    x = MeasurementJsonField(allow_none=True)
    y = MeasurementJsonField(allow_none=True)
    z = MeasurementJsonField(allow_none=True)

    @property
    def load_type(self) -> Type[SamplingLocation]:
        return SamplingLocation


class SampleLocalizationJsonSchema(
    SamplingLocationJsonSchema, LoadingSchema[SamplingLocation]
):
    visual_marking = fields.String(allow_none=True)

    @property
    def load_type(self) -> Type[SamplingLocation]:
        return SampleLocalization


class SamplingConstraintJsonSchema(LoadingSchema[SamplingConstraintJsonModel]):
    """Schema for serializing and deserializing a `SamplingConstraintJsonModel`."""

    identifier = SpecimenIdentifierJsonField()
    sampling_step_index = fields.Integer()

    @property
    def load_type(self) -> Type[SamplingConstraintJsonModel]:
        return SamplingConstraintJsonModel

    @pre_dump
    def dump_simple(self, sampling: Sampling, **kwargs):
        return SamplingConstraintJsonModel.to_json_model(sampling)


class SamplingJsonSchema(DataclassLoadingSchema[SamplingJsonModel]):
    action = fields.Constant(PreparationAction.SAMPLING.value, dump_only=True)
    method = JsonFieldFactory.concept_code(SpecimenSamplingProcedureCode)(
        allow_none=True
    )
    sampling_constraints = fields.List(
        fields.Nested(SamplingConstraintJsonSchema, allow_none=True), allow_none=True
    )
    date_time = fields.DateTime(allow_none=True)
    description = fields.String(allow_none=True)
    location = fields.Nested(SamplingLocationJsonSchema, allow_none=True)

    @property
    def load_type(self) -> Type[SamplingJsonModel]:
        return SamplingJsonModel


class CollectionJsonSchema(DataclassLoadingSchema[Collection]):
    action = fields.Constant(PreparationAction.COLLECTION.value, dump_only=True)
    method = JsonFieldFactory.concept_code(SpecimenCollectionProcedureCode)()
    date_time = fields.DateTime(allow_none=True)
    description = fields.String(allow_none=True)

    @property
    def load_type(self) -> Type[Collection]:
        return Collection


class ProcessingJsonSchema(DataclassLoadingSchema[Processing]):
    action = fields.Constant(PreparationAction.PROCESSING.value, dump_only=True)
    method = JsonFieldFactory.concept_code(SpecimenPreparationStepsCode)(
        allow_none=True
    )
    date_time = fields.DateTime(allow_none=True)
    description = fields.String(allow_none=True)

    @property
    def load_type(self) -> Type[Processing]:
        return Processing


class EmbeddingJsonSchema(DataclassLoadingSchema[Embedding]):
    action = fields.Constant(PreparationAction.EMBEDDING.value, dump_only=True)
    medium = JsonFieldFactory.concept_code(SpecimenEmbeddingMediaCode)()
    date_time = fields.DateTime(allow_none=True)
    description = fields.String(allow_none=True)

    @property
    def load_type(self) -> Type[Embedding]:
        return Embedding


class FixationJsonSchema(DataclassLoadingSchema[Fixation]):
    action = fields.Constant(PreparationAction.FIXATION.value, dump_only=True)
    fixative = JsonFieldFactory.concept_code(SpecimenFixativesCode)()
    date_time = fields.DateTime(allow_none=True)
    description = fields.String(allow_none=True)

    @property
    def load_type(self) -> Type[Fixation]:
        return Fixation


class ReceivingJsonSchema(DataclassLoadingSchema[Receiving]):
    action = fields.Constant(PreparationAction.RECEIVING.value, dump_only=True)
    date_time = fields.DateTime(allow_none=True)
    description = fields.String(allow_none=True)

    @property
    def load_type(self) -> Type[Receiving]:
        return Receiving


class StorageJsonSchema(DataclassLoadingSchema[Storage]):
    action = fields.Constant(PreparationAction.STORAGE.value, dump_only=True)
    date_time = fields.DateTime(allow_none=True)
    description = fields.String(allow_none=True)

    @property
    def load_type(self) -> Type[Storage]:
        return Storage


class PreparationStepJsonSchema(Schema):
    """Schema to use to serialize/deserialize preparation steps."""

    """Mapping step type to schema."""
    _type_to_schema_mapping: Dict[Type[PreparationStep], Type[Schema]] = {
        Sampling: SamplingJsonSchema,
        UnknownSampling: SamplingJsonSchema,
        Collection: CollectionJsonSchema,
        Processing: ProcessingJsonSchema,
        Embedding: EmbeddingJsonSchema,
        Fixation: FixationJsonSchema,
        Receiving: ReceivingJsonSchema,
        Storage: StorageJsonSchema,
    }

    """Mapping key in serialized step to schema."""
    _action_to_schema_mapping: Dict[PreparationAction, Type[Schema]] = {
        PreparationAction.SAMPLING: SamplingJsonSchema,
        PreparationAction.COLLECTION: CollectionJsonSchema,
        PreparationAction.PROCESSING: ProcessingJsonSchema,
        PreparationAction.EMBEDDING: EmbeddingJsonSchema,
        PreparationAction.FIXATION: FixationJsonSchema,
        PreparationAction.RECEIVING: ReceivingJsonSchema,
        PreparationAction.STORAGE: StorageJsonSchema,
    }

    def dump(
        self,
        data: Union[PreparationStep, Iterable[PreparationStep]],
        **kwargs,
    ):
        if isinstance(data, PreparationStep):
            return self._subschema_dump(data)
        return [self._subschema_dump(item) for item in data]

    def load(
        self,
        data: Union[Dict[str, Any], Iterable[Dict[str, Any]]],
        **kwargs,
    ):
        if isinstance(data, dict):
            return self._subschema_load(data)  # type: ignore
        return [self._subschema_load(step) for step in data]

    def _subschema_load(
        self, step: Dict[str, Any]
    ) -> Union[PreparationStep, SamplingJsonModel]:
        """Select a schema and load and return step using the schema."""
        try:
            action = PreparationAction(step.pop("action"))
            schema = self._action_to_schema_mapping[action]
        except KeyError:
            raise NotImplementedError()
        loaded = schema().load(step, many=False)
        assert isinstance(loaded, (PreparationStep, SamplingJsonModel))
        return loaded

    def _subschema_dump(self, step: PreparationStep):
        """Select a schema and dump the step using the schema."""
        schema = self._type_to_schema_mapping[type(step)]
        return schema().dump(step, many=False)


class StainingJsonSchema(DataclassLoadingSchema[Staining]):
    substances = fields.List(
        JsonFieldFactory.concept_code(SpecimenStainsCode)(), allow_none=True
    )
    date_time = fields.DateTime(allow_none=True)
    description = fields.String(allow_none=True)

    @property
    def load_type(self) -> Type[Staining]:
        return Staining


class SpecimenJsonSchema(LoadingSchema[Specimen]):
    """Schema for extracted specimen that has not been sampled from other specimen."""

    identifier = SpecimenIdentifierJsonField()
    steps = fields.List(fields.Nested(PreparationStepJsonSchema()))
    type = JsonFieldFactory.concept_code(AnatomicPathologySpecimenTypesCode)(
        allow_none=True, load_default=None
    )
    container = JsonFieldFactory.concept_code(ContainerTypeCode)(
        allow_none=True, load_default=None
    )

    @property
    def load_type(self) -> Type[SpecimenJsonModel]:
        return SpecimenJsonModel


class SampleJsonSchema(LoadingSchema[SampleJsonModel]):
    """Schema for sampled specimen."""

    identifier = SpecimenIdentifierJsonField()
    steps = fields.List(fields.Nested(PreparationStepJsonSchema()))
    sampled_from = fields.List(fields.Nested(SamplingConstraintJsonSchema))
    type = JsonFieldFactory.concept_code(AnatomicPathologySpecimenTypesCode)(
        allow_none=True, load_default=None
    )
    container = JsonFieldFactory.concept_code(ContainerTypeCode)(
        allow_none=True, load_default=None
    )

    @property
    def load_type(self) -> Type[SampleJsonModel]:
        return SampleJsonModel


class SlideSampleJsonSchema(LoadingSchema[SlideSampleJsonModel]):
    """Schema for sampled specimen on a slide."""

    identifier = SpecimenIdentifierJsonField()
    steps = fields.List(fields.Nested(PreparationStepJsonSchema()))
    anatomical_sites = fields.List(CodeJsonField(), allow_none=True)
    sampled_from = fields.Nested(SamplingConstraintJsonSchema)
    uid = UidJsonField(allow_none=True)
    localization = fields.Nested(
        SampleLocalizationJsonSchema, allow_none=True, load_default=None
    )
    short_description = fields.String(allow_none=True, load_default=None)
    detailed_description = fields.String(allow_none=True, load_default=None)

    @property
    def load_type(self) -> Type[SlideSampleJsonModel]:
        return SlideSampleJsonModel


class BaseSpecimenJsonSchema(Schema):
    """Schema to use to serialize/deserialize specimens."""

    """Mapping specimen type to schema."""
    _type_to_schema_mapping: Dict[Type[BaseSpecimen], Type[Schema]] = {
        Specimen: SpecimenJsonSchema,
        Sample: SampleJsonSchema,
        SlideSample: SlideSampleJsonSchema,
    }

    """Mapping key in serialized specimen to schema."""
    _key_to_schema_mapping: Dict[str, Type[Schema]] = {
        "anatomical_sites": SlideSampleJsonSchema,
        "sampled_from": SampleJsonSchema,
        "type": SpecimenJsonSchema,
    }

    def dump(
        self,
        specimens: Union[BaseSpecimen, Iterable[BaseSpecimen]],
        **kwargs,
    ):
        if isinstance(specimens, BaseSpecimen):
            specimens = [specimens]

        all_specimens: Dict[Union[str, SpecimenIdentifier], BaseSpecimen] = {}
        for specimen in specimens:
            if isinstance(specimen, SampledSpecimen):
                all_specimens.update(self._get_samplings(specimen))

        return [self._subschema_dump(specimen) for specimen in all_specimens.values()]

    def load(
        self,
        data: Union[Dict[str, Any], Iterable[Dict[str, Any]]],
        **kwargs,
    ) -> List[BaseSpecimen]:
        """Load serialized specimen or list of specimen as `Specimen`."""
        if isinstance(data, dict):
            loaded = [self._subschema_load(data)]  # type: ignore
        else:
            loaded = [self._subschema_load(item) for item in data]
        specimen_parser = SpecimenJsonParser(loaded)
        return specimen_parser.create_specimens()

    @classmethod
    def _get_samplings(
        cls, sample: SampledSpecimen
    ) -> Dict[Union[str, SpecimenIdentifier], BaseSpecimen]:
        """Return a dictionary containing this specimen and all recursive sampled specimens."""
        samplings: Dict[Union[str, SpecimenIdentifier], BaseSpecimen] = {
            sample.identifier: sample
        }
        for sampling in sample.sampled_from_list:
            if not isinstance(sampling.specimen, SampledSpecimen):
                samplings.update({sampling.specimen.identifier: sampling.specimen})
            else:
                samplings.update(cls._get_samplings(sampling.specimen))
        return samplings

    @classmethod
    def _subschema_select(cls, specimen: Dict[str, Any]) -> Type[Schema]:
        try:
            return next(
                schema
                for key, schema in cls._key_to_schema_mapping.items()
                if key in specimen
            )
        except StopIteration:
            raise NotImplementedError()

    def _subschema_load(self, specimen: Dict[str, Any]) -> BaseSpecimenJsonModel:
        """Select a schema and load and return specimen using the schema."""
        schema = self._subschema_select(specimen)
        loaded = schema().load(specimen, many=False)
        assert isinstance(loaded, BaseSpecimenJsonModel)
        return loaded

    def _subschema_dump(self, specimen: BaseSpecimen) -> Dict[str, Any]:
        """Select a schema and dump the specimen using the schema."""
        schema = self._type_to_schema_mapping[type(specimen)]
        dumped = schema().dump(specimen)
        assert isinstance(dumped, dict)
        return dumped



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import os
from abc import abstractmethod
from datetime import datetime
from pathlib import Path
from typing import (
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
)

from pydicom.encaps import itemize_frame
from pydicom.tag import ItemTag, SequenceDelimiterTag
from pydicom.uid import UID
from pydicom.valuerep import DSfloat

from wsidicom.codec import Encoder
from wsidicom.errors import WsiDicomBotOverflow
from wsidicom.file.io.frame_index import (
    BotWriter,
    EotWriter,
    OffsetTableType,
    OffsetTableWriter,
)
from wsidicom.file.io.wsidicom_io import WsiDicomIO
from wsidicom.geometry import Point, Region, Size
from wsidicom.instance import ImageData
from wsidicom.instance.dataset import WsiDataset
from wsidicom.tags import LossyImageCompressionRatioTag, PixelDataTag
from wsidicom.thread import ConditionalThreadPoolExecutor
from wsidicom.uid import WSI_SOP_CLASS_UID


class WsiDicomWriter:
    """Writer for DICOM WSI files."""

    def __init__(
        self, file: WsiDicomIO, transfer_syntax: UID, offset_table: OffsetTableType
    ) -> None:
        """
        Create a writer for DICOM WSI data.

        Parameters
        ----------
        file: WsiDicomIO
            File to open for writing.
        transfer_syntax: UID
            Transfer syntax to use.
        offset_table: OffsetTableType
            Offset table to use.
        """
        if not self.supports_transfer_syntax(transfer_syntax):
            raise ValueError(
                f"Transfer syntax not supported for writer of type {type(self)}."
            )
        if not self.supports_offset_table(offset_table):
            raise ValueError(
                f"Offset table not supported for writer of type {type(self)}."
            )
        self._file = file
        self._file.is_implicit_VR = transfer_syntax.is_implicit_VR
        self._file.is_little_endian = transfer_syntax.is_little_endian
        self._transfer_syntax = transfer_syntax
        self._offset_table = offset_table

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @classmethod
    def open(
        cls, file: Path, transfer_syntax: UID, offset_table: OffsetTableType
    ) -> "WsiDicomWriter":
        """Open file in path as WsiDicomWriter.

        Parameters
        ----------
        file: Path
            Path to file.
        transfer_syntax: UID
            Transfer syntax to use.
        offset_table: OffsetTableType
            Offset table to use.

        Returns
        -------
        WsiDicomWriter
            WsiDicomWriter for file.
        """
        stream = WsiDicomIO.open(file, "w+b")
        if transfer_syntax.is_encapsulated:
            writer = WsiDicomEncapsulatedWriter
        else:
            writer = WsiDicomNativeWriter
        return writer(stream, transfer_syntax, offset_table)

    def write(
        self,
        uid: UID,
        dataset: WsiDataset,
        data: Dict[Tuple[str, float], ImageData],
        workers: int,
        chunk_size: int,
        instance_number: int,
        scale: int = 1,
        transcoder: Optional[Encoder] = None,
    ) -> None:
        """
        Write data to file.

        Parameters
        ----------
        uid: UID
            Instance UID for file.
        transfer_syntax: UID.
            Transfer syntax for file
        dataset: WsiDataset
            Dataset to write (excluding pixel data).
        data: Dict[Tuple[str, float], ImageData]
            Pixel data to write.
        workers: int
            Number of workers to use for writing pixel data.
        chunk_size: int
            Number of frames to give each worker.
        offset_table: OffsetTableType
            Offset table to use.
        instance_number: int
            Instance number for file.
        scale: int = 1
            Scale factor.
        transcoder: Optional[Encoder] = None,
            Encoder to use if transcoding pixel data. Default is None to not transcode.

        """
        if transcoder is not None:
            if transcoder.transfer_syntax != self._transfer_syntax:
                raise ValueError("Transcoder transfer syntax must match writer.")
        self._file.write_preamble()
        self._file.write_file_meta_info(uid, WSI_SOP_CLASS_UID, self._transfer_syntax)
        dataset.SOPInstanceUID = uid
        dataset.InstanceNumber = instance_number
        dataset_start = self._file.tell()
        self._file.write_dataset(dataset, datetime.now())
        dataset_end = self._file.tell()
        self._write_pixel_data(
            data,
            dataset,
            (dataset_start, dataset_end),
            workers,
            chunk_size,
            scale,
            transcoder,
        )

    @abstractmethod
    def _write_pixel_data(
        self,
        data: Dict[Tuple[str, float], ImageData],
        dataset: WsiDataset,
        dataset_range: Tuple[int, int],
        workers: int,
        chunk_size: int,
        scale: int,
        transcoder: Optional[Encoder],
    ) -> List[int]:
        """Write pixel data to file.

        Parameters
        ----------
        data: Dict[Tuple[str, float], ImageData]
            Pixel data to write.
        dataset: WsiDataset
            Dataset with parameters for image to write.
        dataset_start: int
            Position of dataset in file.
        workers: int
            Number of workers to use for writing pixel data.
        chunk_size: int
            Number of frames to give each worker.
        scale: int
            Scale factor. Set to 1 for no scaling.
        transcoder: Optional[Encoder]
            Encoder to use if transcoding image.
        """
        raise NotImplementedError()

    @abstractmethod
    def _write_tile(self, tile: bytes) -> int:
        """Write tile to file and return position of frame in file."""
        raise NotImplementedError()

    @abstractmethod
    def supports_transfer_syntax(self, transfer_syntax: UID) -> bool:
        """Return True if writer supports transfer syntax."""
        raise NotImplementedError()

    @abstractmethod
    def supports_offset_table(self, offset_table: OffsetTableType) -> bool:
        """Return True if writer supports offset table."""
        raise NotImplementedError()

    def close(self, force: Optional[bool] = False) -> None:
        self._file.close(force)

    def _write_tiles(
        self,
        image_data: ImageData,
        z: float,
        path: str,
        workers: int,
        chunk_size: int,
        scale: int = 1,
        transcoder: Optional[Encoder] = None,
    ) -> List[int]:
        """Write pixel data to file.

        Parameters
        ----------
        image_data: ImageData
            Image data to read pixel tiles from.
        encapsulate: bool
            If pixel data should be encapsulated.
        z: float
            Focal plane to write.
        path: str
            Optical path to write.
        workers: int
            Maximum number of thread workers to use.
        chunk_size: int
            Chunk size (number of tiles) to process at a time. Actual chunk
            size also depends on minimun_chunk_size from image_data.
        scale: int
            Scale factor (1 = No scaling).

        Returns
        -------
        List[int]
            List of frame position (position of ItemTag), relative to start of
            file.
        """
        chunked_tile_points = self._chunk_tile_points(image_data, chunk_size, scale)

        def get_tiles(tile_points: Iterable[Point]) -> List[bytes]:
            """Function to get tiles as bytes."""
            return list(
                image_data.get_encoded_tiles(tile_points, z, path, scale, transcoder)
            )

        with ConditionalThreadPoolExecutor(max_workers=workers) as pool:
            return [
                self._write_tile(tile)
                for thread_result in pool.map(get_tiles, chunked_tile_points)
                for tile in thread_result
            ]

    def _chunk_tile_points(
        self, image_data: ImageData, chunk_size: int, scale: int = 1
    ) -> Iterator[Iterator[Point]]:
        """Divide tile positions in image_data into chunks.

        Parameters
        ----------
        image_data: ImageData
            Image data with tiles to chunk.
        chunk_size: int
            Requested chunk size
        scale: int = 1
            Scaling factor (1 = no scaling).

        Returns
        -------
        Iterator[Iterator[Point]]
            Chunked tile positions
        """
        minimum_chunk_size = getattr(image_data, "suggested_minimum_chunk_size", 1)
        # If chunk_size is less than minimum_chunk_size, use minimum_chunk_size
        # Otherwise, set chunk_size to highest even multiple of
        # minimum_chunk_size
        chunk_size = max(
            minimum_chunk_size, chunk_size // minimum_chunk_size * minimum_chunk_size
        )
        new_tiled_size = image_data.tiled_size.ceil_div(scale)
        # Divide the image tiles up into chunk_size chunks (up to tiled size)
        chunked_tile_points = (
            Region(
                Point(x, y), Size(min(chunk_size, new_tiled_size.width - x), 1)
            ).iterate_all()
            for y in range(new_tiled_size.height)
            for x in range(0, new_tiled_size.width, chunk_size)
        )
        return chunked_tile_points


class WsiDicomEncapsulatedWriter(WsiDicomWriter):
    @classmethod
    def open(
        cls, file: Path, transfer_syntax: UID, offset_table: OffsetTableType
    ) -> "WsiDicomEncapsulatedWriter":
        """Open file in path as WsiDicomEncapsulatedWriter.

        Parameters
        ----------
        file: Path
            Path to file.
        transfer_syntax: UID
            Transfer syntax to use.
        offset_table: OffsetTableType
            Offset table to use.

        Returns
        -------
        WsiDicomEncapsulatedWriter
            WsiDicomEncapsulatedWriter for file.
        """
        stream = WsiDicomIO.open(file, "w+b")
        return cls(stream, transfer_syntax, offset_table)

    def copy(
        self,
        copy_from: WsiDicomIO,
        dataset_end: int,
        pixels_end: int,
        frame_positions: List[int],
    ) -> List[int]:
        """Copy dataset and pixel data from other file to this.

        Parameters
        ----------
        copy_from: DicomFileLike
            File to copy from. Must have encapsulated pixel data.
        dataset_end: int
            Position of EOT or PixelData tag in copy_from.
        pixels_end: int
            End of PixelData in copy_from.
        frame_positions: List[int]
            List of frame positions in copy_from, relative to start of file.

        Returns
        -------
        List[int]
            List of frame position relative to start of new file.
        """
        # Copy dataset until EOT or PixelData tag
        copy_from.seek(0)
        self._file.write(copy_from.read(dataset_end))

        # Write new pixel data start
        (
            new_pixels_start,
            table_writer,
        ) = self._write_pixel_data_start(len(frame_positions))

        # Copy pixel data
        first_frame_position = frame_positions[0]
        copy_from.seek(first_frame_position)
        while copy_from.tell() < pixels_end:
            chunk_size = min(4096, pixels_end - copy_from.tell())
            self._file.write(copy_from.read(chunk_size))

        # Adjust frame positions
        frame_position_change = new_pixels_start - first_frame_position
        new_frame_positions = [
            position + frame_position_change for position in frame_positions
        ]
        last_frame_end = self._file.tell()

        # Write pixel data end and EOT or BOT if used.
        frame_positions = self._write_pixel_data_end(
            table_writer,
            new_pixels_start,
            last_frame_end,
            dataset_end,
            new_frame_positions,
        )
        return frame_positions

    def _write_pixel_data(
        self,
        data: Dict[Tuple[str, float], ImageData],
        dataset: WsiDataset,
        dataset_range: Tuple[int, int],
        workers: int,
        chunk_size: int,
        scale: int,
        transcoder: Optional[Encoder],
    ) -> List[int]:
        (
            pixels_start,
            table_writer,
        ) = self._write_pixel_data_start(dataset.frame_count)
        frame_positions = [
            position
            for (path, z), image_data in sorted(data.items())
            for position in self._write_tiles(
                image_data, z, path, workers, chunk_size, scale, transcoder
            )
        ]
        last_frame_end = self._file.tell()
        if transcoder is not None and transcoder.lossy:
            compressed_size = self._calculate_size(frame_positions, last_frame_end)
            self._set_compression_ratio(dataset_range[0], dataset, compressed_size)

        frame_positions = self._write_pixel_data_end(
            table_writer,
            pixels_start,
            last_frame_end,
            dataset_range[1],
            frame_positions,
        )

        return frame_positions

    def _write_tile(
        self,
        tile: bytes,
    ) -> int:
        position = self._file.tell()
        for frame in itemize_frame(tile, 1):
            self._file.write(frame)
        return position

    def supports_transfer_syntax(self, transfer_syntax: UID) -> bool:
        return transfer_syntax.is_encapsulated

    def supports_offset_table(self, offset_table: OffsetTableType) -> bool:
        return offset_table is not OffsetTableType.NONE

    def _write_pixel_data_start(
        self,
        number_of_frames: int,
    ) -> Tuple[int, Optional[OffsetTableWriter]]:
        """Write tags starting pixel data and reserves space for BOT or EOT.

        Parameters
        ----------
        number_of_frames: int
            Number of frames to reserve space for in BOT or EOT.

        Returns
        -------
        Tuple[int, Optional[OffsetTableWriter]]:
            Start of pixel data (after BOT) and optional offset table writer.
        """
        table_writer = None
        if self._offset_table == OffsetTableType.EXTENDED:
            table_writer = EotWriter(self._file)
            table_writer.reserve(number_of_frames)

        # Write pixel data tag
        self._file.write_tag_of_vr_and_length(PixelDataTag, "OB")

        if self._offset_table == OffsetTableType.BASIC:
            table_writer = BotWriter(self._file)
            table_writer.reserve(number_of_frames)
        elif (
            self._offset_table == OffsetTableType.EMPTY
            or self._offset_table == OffsetTableType.EXTENDED
        ):
            self._file.write_tag(ItemTag)
            self._file.write_leUL(0)

        pixel_data_start = self._file.tell()

        return pixel_data_start, table_writer

    def _write_pixel_data_end(
        self,
        offset_writer: Optional[OffsetTableWriter],
        pixels_start: int,
        last_frame_end: int,
        dataset_end: int,
        frame_positions: List[int],
    ):
        """Write tags ending pixel data and BOT or EOT.

        Parameters
        ----------
        offset_writer: Optional[OffsetTableWriter]
            Offset table writer to use.
        pixels_start: int
            Position of start of pixel data.
        dataset_end: int
            Position of EOT or PixelData tag.
        frame_positions: List[int]
            List of frame positions in file, relative to start of file.
        """
        self._write_pixel_data_end_tag()
        if offset_writer is not None:
            try:
                offset_writer.write(pixels_start, frame_positions, last_frame_end)
            except WsiDicomBotOverflow as exception:
                if self._file.filepath is not None and isinstance(
                    offset_writer, BotWriter
                ):
                    frame_positions = self._rewrite_as_table(
                        OffsetTableType.EXTENDED,
                        dataset_end,
                        last_frame_end,
                        frame_positions,
                    )
                else:
                    raise exception
        return frame_positions

    def _write_pixel_data_end_tag(self) -> None:
        """Writes tags ending pixel data."""
        self._file.write_tag(SequenceDelimiterTag)
        self._file.write_leUL(0)

    def _rewrite_as_table(
        self,
        offset_table: OffsetTableType,
        dataset_end: int,
        pixels_end: int,
        frame_positions: List[int],
    ) -> List[int]:
        """Rewrite file as encapsulated with EOT. Closes current file and replaces
        it with the new a new file.

        Parameters
        ----------
        offset_table: OffsetTableType
            Offset table to use in new file.
        dataset_end: int
            Position of EOT or PixelData tag in current file.
        pixels_end: int
            End of PixelData in current file.
        frame_positions: List[int]
            List of frame positions in current file, relative to start of file.

        Returns
        -------
        List[int]
            List of frame position relative to start of new file.
        """
        if self._file.filepath is None:
            raise ValueError("Cannot rewrite file without filepath.")
        temp_file_path = self._file.filepath.with_suffix(".tmp")
        with WsiDicomEncapsulatedWriter.open(
            temp_file_path, self._transfer_syntax, offset_table
        ) as writer:
            frame_positions = writer.copy(
                self._file,
                dataset_end,
                pixels_end,
                frame_positions,
            )
        self._file.close()
        os.replace(temp_file_path, self._file.filepath)
        self._file = WsiDicomIO.open(self._file.filepath, "r+b")
        return frame_positions

    @staticmethod
    def _calculate_size(
        frame_positions: Sequence[int], last_frame_position: int
    ) -> int:
        first_frame_start = frame_positions[0]
        last_frame_end = last_frame_position
        return last_frame_end - first_frame_start - len(frame_positions) * 16

    def _set_compression_ratio(
        self, dataset_start: int, dataset: WsiDataset, compressed_size: int
    ):
        uncompressed_size = (
            dataset.frame_count
            * dataset.tile_size.area
            * dataset.samples_per_pixel
            * (dataset.bits // 8)
        )
        ratio = DSfloat(round(uncompressed_size / compressed_size, 2))
        ratios = dataset.get_multi_value(LossyImageCompressionRatioTag)
        ratios[-1] = ratio
        self._file.update_dataset(
            dataset_start,
            {LossyImageCompressionRatioTag: ratios},
        )


class WsiDicomNativeWriter(WsiDicomWriter):
    @classmethod
    def open(
        cls, file: Path, transfer_syntax: UID, offset_table: OffsetTableType
    ) -> "WsiDicomNativeWriter":
        """Open file in path as WsiDicomNativeWriter.

        Parameters
        ----------
        file: Path
            Path to file.
        transfer_syntax: UID
            Transfer syntax to use.
        offset_table: OffsetTableType
            Offset table to use.

        Returns
        -------
        WsiDicomNativeWriter
            WsiDicomNativeWriter for file.
        """
        stream = WsiDicomIO.open(file, "w+b")
        return cls(stream, transfer_syntax, offset_table)

    def _write_pixel_data(
        self,
        data: Dict[Tuple[str, float], ImageData],
        dataset: WsiDataset,
        dataset_range: Tuple[int, int],
        workers: int,
        chunk_size: int,
        scale: int,
        transcoder: Optional[Encoder],
    ) -> List[int]:
        length = (
            dataset.tile_size.area
            * dataset.samples_per_pixel
            * (dataset.bits // 8)
            * dataset.frame_count
        )
        self._file.write_tag_of_vr_and_length(PixelDataTag, "OB", length)
        return [
            position
            for (path, z), image_data in sorted(data.items())
            for position in self._write_tiles(
                image_data, z, path, workers, chunk_size, scale, transcoder
            )
        ]

    def _write_tile(self, tile: bytes) -> int:
        position = self._file.tell()
        self._file.write(tile)
        return position

    def supports_transfer_syntax(self, transfer_syntax: UID) -> bool:
        return not transfer_syntax.is_encapsulated

    def supports_offset_table(self, offset_table: OffsetTableType) -> bool:
        return offset_table is OffsetTableType.NONE



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module with DICOM specimen description formatter."""

from typing import (
    Iterable,
    Iterator,
    Optional,
    Sequence,
)

from wsidicom.metadata.sample import (
    BaseSampling,
    BaseSpecimen,
    SampledSpecimen,
    Sampling,
    SlideSample,
    SpecimenIdentifier,
    Staining,
)
from wsidicom.metadata.schema.dicom.sample.model import (
    SamplingDicomModel,
    SpecimenDescriptionDicomModel,
    SpecimenPreparationStepDicomModel,
    StainingDicomModel,
)


class SpecimenDicomFormatter:
    """Formatter for producing DICOM Specimen Description models from slide samples.

    Converting a slide sample to a DICOM Specimen Description requires producing a
    list of all the preparation steps that were performed on the sample. As a parent
    to a sample may have been sampled from multiple specimens, we use the sampling
    constraints to limit the sampling tree to a single branch."""

    @classmethod
    def to_dicom(
        cls,
        slide_sample: SlideSample,
        stains: Optional[Sequence[Staining]] = None,
    ) -> SpecimenDescriptionDicomModel:
        """Create a DICOM specimen description for slide sample.

        Parameters
        ----------
        slide_sample : SlideSample
            The slide sample to convert to DICOM.
        stains : Optional[Sequence[Staining]], optional
            Stainings performed on the slide sample.

        Returns
        -------
        SpecimenDescriptionDicomModel
            DICOM model describing the slide sample.
        """
        if stains is None:
            stains = []
        if slide_sample.uid is None:
            sample_uid = slide_sample.default_uid
        else:
            sample_uid = slide_sample.uid
        sample_preparation_steps = cls._get_steps(slide_sample, stains)
        identifier, issuer = SpecimenIdentifier.get_identifier_and_issuer(
            slide_sample.identifier
        )
        return SpecimenDescriptionDicomModel(
            identifier=identifier,
            issuer_of_identifier=issuer,
            uid=sample_uid,
            steps=list(sample_preparation_steps),
            anatomical_sites=list(slide_sample.anatomical_sites)
            if slide_sample.anatomical_sites is not None
            else [],
            short_description=slide_sample.short_description,
            detailed_description=slide_sample.detailed_description,
            localization=slide_sample.localization,
        )

    @classmethod
    def _get_steps(
        cls, slide_sample: SlideSample, stainings: Optional[Iterable[Staining]]
    ) -> Iterator[SpecimenPreparationStepDicomModel]:
        """Get all the steps performed on the slide sample, following the sampling
        branch specified in the sampling constraints.

        Parameters
        ----------
        slide_sample : SlideSample
            The slide sample to get steps for
        stainings: Optional[Iterable[Staining]]
            Stainings performed on the slide sample.

        Returns
        -------
        Iterator[SpecimenPreparationStepDicomModel]
            Iterator of DICOM steps performed on the slide sample.
        """
        yield from (
            step
            for sampling in slide_sample.sampled_from_list
            for step in cls._get_steps_for_sampling(sampling, slide_sample)
        )
        if stainings is not None:
            for staining in stainings:
                yield StainingDicomModel.from_step(staining, slide_sample)

    @classmethod
    def _get_steps_for_sampling(
        cls,
        sampling: BaseSampling,
        sample: BaseSpecimen,
    ) -> Iterator[SpecimenPreparationStepDicomModel]:
        """Return DICOM steps for the specimen the sample was sampled from."""
        if isinstance(sampling.specimen, SampledSpecimen):
            yield from (
                step
                for sub_sampling in sampling.specimen.sampled_from_list
                if sampling.sampling_constraints is None
                or sub_sampling in sampling.sampling_constraints
                for step in cls._get_steps_for_sampling(sub_sampling, sampling.specimen)
            )

        yield from (
            dicom_step
            for dicom_step in [
                SpecimenPreparationStepDicomModel.from_step(step, sampling.specimen)
                for step in sampling.specimen.steps
                if not isinstance(step, BaseSampling)
            ]
            if dicom_step is not None
        )
        if isinstance(sampling, Sampling):
            yield SamplingDicomModel.from_step(sampling, sample)
